import time
import numpy as np
from starred.deconvolution.deconvolution import Deconv
from starred.psf.psf import PSF
from starred.utils.jax_utils import decompose
from starred.utils.generic_utils import Upsample
from copy import deepcopy
from scipy.ndimage import shift
import matplotlib.pyplot as plt
import warnings

def propagate_noise(model, noise_maps, kwargs, masks=None, wavelet_type_list=['starlet'],
                    method='MC', num_samples=100, seed=None,
                    likelihood_type='chi2', verbose=False,
                    upsampling_factor=1, scaling_noise_ref=None):
    """Performs noise propagation using MC or SLIT methods.
    
    :param model: array containing the model
    :param noise_maps: array containing the noise maps
    :param kwargs:  dictionary containing the parameters of the model
    :param masks: array containing the masks
    :param wavelet_type_list: list with the appropriate wavelet families
    :param method: method for noise propagation. Choose 'MC' for an empirical propagation of the noise or 'SLIT' for analytical propagation.
    :param num_samples: number of realizations for the MC method. Not used if `method`='SLIT'.
    :type num_samples: int
    :param upsampling_factor: the upsampling factor
    :type upsampling_factor: int
    :param scaling_noise_ref: index position of the reference noise map (lambda is given in the unit of standard deviation for this noise map).
    Leave it to `None` to take the mean of all provided noise maps (recommended). This only works for the PSF reconstruction. For the deconvolution,
    the reference image's noise map is always used, so lambda is given in unit of noise level, according to this noise map.
    :type scaling_noise_ref: integer

    """
    if likelihood_type not in ['l2_norm', 'chi2']:
        raise ValueError("Only 'l2_norm' and 'chi2' are supported options for likelihood_type.")

    n_image, n_pix, n_pix = np.shape(noise_maps)
    if isinstance(model, PSF):
        if scaling_noise_ref is None:
            masked_noise = deepcopy(noise_maps)
            if masks is not None:
                ind = np.where(masks == 0)
                masked_noise = deepcopy(noise_maps)
                masked_noise[ind] = np.nan

            centered_masked_noise_maps = np.zeros_like(masked_noise)
            #center the noise maps according to the estimated shift, we just use linear interpolation here
            for i in range(n_image):
                centered_masked_noise_maps[i] = shift(deepcopy(masked_noise[i]), (
                - kwargs['kwargs_gaussian']['y0'][i],
                - kwargs['kwargs_gaussian']['x0'][i]), mode='constant', cval=np.nan, order=1)
            noise_map = np.nanmean(centered_masked_noise_maps, axis=0)

        else:
            if masks is not None:
                if not np.all(masks[scaling_noise_ref, :, :] == 1.):
                    raise RuntimeError(
                        "Your reference star contain masked pixels, I cannot propagate the noise correctly. Use `scaling_noise_ref = None` or remove the mask.")
            noise_map = shift(deepcopy(noise_maps[scaling_noise_ref, :, :]), (
                - kwargs['kwargs_gaussian']['y0'][scaling_noise_ref],
                - kwargs['kwargs_gaussian']['x0'][scaling_noise_ref]), mode='constant', cval=np.nan, order=1) #x0 and y0 are inverted

        #check that there is no region which is not covered by any stars. If we are unlucky and this is the case,
        # we will fill with the median noise level.
        noise_map = np.nan_to_num(noise_map, nan=np.nanmedian(noise_map)) / np.sqrt(
            n_image)  # effective noise maps, scale as 1/sqrt(number of stars)

        #get the centered convolution kernel
        kwargs_centered = deepcopy(kwargs['kwargs_gaussian'])
        kwargs_centered['x0'] = np.zeros(n_image)
        kwargs_centered['y0'] = np.zeros(n_image)
        deconv_kernel = model.get_gaussian(0, kwargs_centered, deepcopy(
            kwargs['kwargs_moffat']))  # get the gaussian kernel, will be normalised to 1 later
        a = 1 # leave the possibility to adjust but this should be set to one

    elif isinstance(model, Deconv):
        # for the deconvolution, things are simpler, the reference image always is used to scale the lambda parameter.
        if scaling_noise_ref is not None:
            if scaling_noise_ref != 0:
                warnings.warn(
                    'For the deconvolution, we will use the reference image for the noise propagation in any cases. '
                    '"scaling_noise_ref" is not used.')
        noise_map = deepcopy(noise_maps[0, :, :]) / np.sqrt(
            n_image)  # effective noise maps, assuming all images have the same exposure time
        deconv_kernel = model.psf[0, :, :]
        a = 1  # leave the possibility to adjust but this should be set to one
    else:
        raise TypeError('Unknown instance.')

    deconv_kernel /= np.sum(deconv_kernel)  # normalise kernel

    var_map = np.array(noise_map ** 2)  # cast to numpy array otherwise computations are slowed down
    std_map = np.array(noise_map)

    # wavelet transform Phi^T operators
    PhiT_operator_list = []
    num_scales_list = []
    for wavelet_type in wavelet_type_list:
        if wavelet_type in ['battle-lemarie-1', 'battle-lemarie-3']:
            num_scales = 1  # we only care about the first scale for this one
        else:
            num_scales = int(np.log2(min(n_pix * upsampling_factor, n_pix * upsampling_factor)))
        # wavelet = jax_utils.WaveletTransform(num_scales,
        #                                      wavelet_type=wavelet_type,
        #                                      )
        PhiT_operator_list.append(decompose)
        num_scales_list.append(num_scales)

    nx_psi, ny_psi = n_pix * upsampling_factor, n_pix * upsampling_factor

    psi_wt_std_list = []
    # map noise values to source plane

    if method == 'MC':
        psi_wt_std_list = []
        for wavelet_type, PhiT_operator, num_scales in zip(wavelet_type_list, PhiT_operator_list, num_scales_list):
            start = time.time()
            psi_wt_reals = []
            np.random.seed(seed)
            std_map_up = Upsample(std_map, factor=upsampling_factor)
            var_map_up = Upsample(var_map, factor=upsampling_factor)
            for i in range(num_samples):
                noise_i = std_map_up * np.random.randn(*std_map_up.shape)  # draw a noise realization
                noise_i2 = std_map_up * np.random.randn(*std_map_up.shape)  # draw a noise realization

                # if chi2 loss, rescale by the data variance
                if likelihood_type == 'chi2':
                    noise_i /= var_map_up  # operator Ck -1 (so that w has unit of 1/flux)

                # before, noise_i was upscaled here and the result stored in psi_i.
                # Now the upscaling is done earlier, but we keep the link psi_i = noise_i
                psi_i = noise_i
                psi_i_blurred = model._convolve(psi_i, deconv_kernel)  # Bt operator
                psi_i_blurred *= a  # transpose of scalar multiplication is the multiplication by a scalar
                psi_wt_i = PhiT_operator(psi_i_blurred, num_scales)
                psi_wt_reals.append(psi_wt_i)

            psi_wt_reals = np.array(psi_wt_reals)  # --> shape = (num_samples, num_scales, nx_psi, ny_psi)
            if verbose: print(f"loop over MC samples for wavelet '{wavelet_type}':", time.time() - start)

            # compute the variance per wavelet scale per potential pixel over all the samples
            psi_wt_var = np.var(psi_wt_reals, axis=0)
            # check
            if np.any(psi_wt_var < 0.):  # pragma: no cover
                raise ValueError("Negative variance terms!")

            # convert to standard deviation
            psi_wt_std = np.sqrt(psi_wt_var)
            psi_wt_std_list.append(psi_wt_std)

    elif method == 'SLIT':

        for PhiT_operator, num_scales in zip(PhiT_operator_list, num_scales_list):

            noise_diag = std_map * np.sqrt(np.sum(deconv_kernel.T ** 2))
            if likelihood_type == 'chi2':
                noise_diag /= var_map
            noise_diag_up = Upsample(noise_diag, factor=upsampling_factor)

            blurred_noise = a * model._convolve(noise_diag_up, deconv_kernel)

            dirac = np.zeros((nx_psi, ny_psi))
            dirac[nx_psi // 2, ny_psi // 2] = 1
            dirac_wt = PhiT_operator(dirac, num_scales)

            psi_wt_std = []
            for k in range(num_scales + 1):
                psi_wt_std2_k = model._convolve(blurred_noise ** 2, dirac_wt[k] ** 2)
                psi_wt_std_k = np.sqrt(psi_wt_std2_k)
                psi_wt_std_k = np.nan_to_num(psi_wt_std_k, nan=np.nanmin(
                    psi_wt_std_k)) * upsampling_factor ** 2  # scaling is important here to match 'MC' scaling convention
                psi_wt_std.append(psi_wt_std_k)
            psi_wt_std = np.array(psi_wt_std)  # --> shape = (num_scales, nx_psi, ny_psi)
            psi_wt_std_list.append(psi_wt_std)  # one per type of (wavelet, num_scales)
    else:
        raise ValueError(f"Method '{method}' for noise propagation is not supported.")

    return psi_wt_std_list


def dirac_impulse(num_pix):
    """
    Returns a 2D array with a Dirac impulse at its center.

    :param num_pix: number of pixels per axis
    :type num_pix: int
    :return: 2D array

    """
    dirac = np.zeros((num_pix, num_pix), dtype=float)
    dirac[int(num_pix / 2), int(num_pix / 2)] = 1.
    return dirac



from copy import deepcopy

import jax.numpy as jnp
import numpy as np

__all__ = ['Parameters']


class Parameters(object):
    """
    Parameters class.

    """

    def __init__(self, kwargs_init, kwargs_fixed, kwargs_up=None, kwargs_down=None):
        """
        :param kwargs_init: dictionary with information on the initial values of the parameters
        :param kwargs_fixed: dictionary containing the fixed parameters 
        :param kwargs_up: dictionary with information on the upper bounds of the parameters 
        :param kwargs_down: dictionary with information on the lower bounds of the parameters 

        """
        self._kwargs_init = self.convert_in_jnp_array(kwargs_init)
        self._kwargs_fixed = self.convert_in_jnp_array(kwargs_fixed)
        if kwargs_up is not None:
            self._kwargs_up = self.convert_in_jnp_array(kwargs_up)
        else:
            self._kwargs_up = None
        if kwargs_down is not None:
            self._kwargs_down = self.convert_in_jnp_array(kwargs_down)
        else:
            self._kwargs_down = None
        self._kwargs_free_indices = self.get_nan_indices()
        
        # update the bounds in case we only have sub parts of our parameter-arrays
        # that are free:
        uku, ukd = self._update_bounds(self._kwargs_init, self._kwargs_fixed, self._kwargs_up, self._kwargs_down)
        self._kwargs_up = uku
        self._kwargs_down = ukd
            

    @property
    def optimized(self):
        """Checks whether a function is optimized."""
        return hasattr(self, '_map_values')

    def convert_in_jnp_array(self, kwargs):
        new_kwargs = {}
        for key in kwargs.keys():
            new_kwargs[key] = {}
            for key2 in kwargs[key].keys():
                new_kwargs[key][key2] = jnp.asarray(deepcopy(kwargs[key][key2]))

        return new_kwargs

    def initial_values(self, as_kwargs=False, copy=False):
        """Returns the initial values of the parameters."""
        if as_kwargs:
            return deepcopy(self._kwargs_init) if copy else self._kwargs_init
        else:
            return deepcopy(self._init_values) if copy else self._init_values

    def current_values(self, as_kwargs=False, restart=False, copy=False):
        """Returns the current values of the parameters."""
        if restart is True or not self.optimized:
            return self.initial_values(as_kwargs=as_kwargs, copy=copy)
        return self.best_fit_values(as_kwargs=as_kwargs, copy=copy)

    def best_fit_values(self, as_kwargs=False, copy=False):
        """Maximum-a-postriori estimate."""
        if as_kwargs:
            return deepcopy(self._kwargs_map) if copy else self._kwargs_map
        else:
            return deepcopy(self._map_values) if copy else self._map_values

    def set_best_fit(self, args):
        """Sets the maximum-a-postriori estimate as the parameter values."""
        self._map_values = args
        self._kwargs_map = self.args2kwargs(self._map_values)

    def _update_arrays(self):
        self._init_values = self.kwargs2args(self._kwargs_init)
        self._kwargs_init = self.args2kwargs(self._init_values)  # for updating missing fields
        self._num_params = len(self._init_values)
        if self.optimized:
            self._map_values = self.kwargs2args(self._kwargs_map)

    def _set_params(self, kwargs, kwargs_key):
        """Setting the parameters."""
        args = []
        kwargs_profile = kwargs[kwargs_key]
        kwargs_fixed_k = self._kwargs_fixed[kwargs_key]
        param_names = self.get_param_names_for_model(kwargs_key)
        for name in param_names:
            if name not in kwargs_fixed_k:
                if isinstance(kwargs_profile[name], list):
                    args += kwargs_profile[name]
                elif isinstance(kwargs_profile[name], (np.ndarray, np.generic, jnp.ndarray, jnp.generic)):
                    el = kwargs_profile[name].tolist()
                    if hasattr(el, '__len__'):
                        args += el
                    else:
                        args += [el]
                else:
                    args += [kwargs_profile[name]]
            else:
                # add the indices in the self._kwargs_free_indices
                free_ind = self._kwargs_free_indices[kwargs_key][name]
                if len(free_ind) >= 1:
                    args += kwargs_profile[name][free_ind].tolist()
        return args
    
    @staticmethod
    def _update_bounds(kwargs_init, kwargs_fixed, kwargs_up, kwargs_down):
        """
        Called during initialization of this class.
        Updates the bounds (kwargs_up and kwargs_down) based on the fixed parameters (kwargs_fixed).
        NaN values in kwargs_fixed represent parameters that are still free to be optimized.
        The function adjusts the bounds to include only the indices where there are NaN values in kwargs_fixed.
    
        :param kwargs_init: Initial parameters
        :param kwargs_fixed: Fixed parameters, with NaN for values that are still optimized
        :param kwargs_up: Upper bounds for parameters
        :param kwargs_down: Lower bounds for parameters
        :return: Updated kwargs_up and kwargs_down
        """
        for main_key in kwargs_fixed:
            for sub_key in kwargs_fixed[main_key]:
                fixed_values = np.array(kwargs_fixed[main_key][sub_key])  # Ensuring it's a numpy array
                if isinstance(fixed_values, np.ndarray):
                    nan_indices = np.isnan(fixed_values)
                    if np.any(nan_indices):
                        if kwargs_up is not None:
                            if not np.sum(nan_indices) == len(
                                    kwargs_up[main_key][sub_key]):  # check if the kwargs have already been updated
                                kwargs_up[main_key][sub_key] = np.array(kwargs_up[main_key][sub_key])[nan_indices]
                        if kwargs_down is not None:
                            if not np.sum(nan_indices) == len(
                                    kwargs_down[main_key][sub_key]):  # check if the kwargs have already been updated
                                kwargs_down[main_key][sub_key] = np.array(kwargs_down[main_key][sub_key])[nan_indices]
        return kwargs_up, kwargs_down

    def get_bounds(self):
        """Returns the upper and lower bounds of the parameters."""
        if self._kwargs_up is None or self._kwargs_down is None:
            return None
        else:
            list_down_limit = []
            list_up_limit = []
            for kwargs_key in self._kwargs_down.keys():
                param_names = self.get_param_names_for_model(kwargs_key)
                for name in param_names:
                    if not name in self._kwargs_fixed[kwargs_key].keys():
                        assert name in self._kwargs_up[kwargs_key].keys(), \
                            "Missing '%s' key in the kwargs_up['%s']" % (name, kwargs_key)
                        assert name in self._kwargs_down[
                            kwargs_key].keys(), "Missing '%s' key in the kwargs_down['%s']" % (name, kwargs_key)
                        up = self._kwargs_up[kwargs_key][name]
                        down = self._kwargs_down[kwargs_key][name]

                    else:
                        # even if key in fixed, maybe we have nans - meaning
                        # a subset is not fixed.
                        free = self._kwargs_free_indices[kwargs_key][name]
                        if len(free) == 0:
                            continue
                        up = self._kwargs_up[kwargs_key][name]
                        down = self._kwargs_down[kwargs_key][name]
                        
                    if isinstance(down, list):
                        list_down_limit += down
                    elif isinstance(down, (np.ndarray, jnp.ndarray)):
                        el = down.tolist()
                        if hasattr(el, '__len__'):
                            list_down_limit += el
                        else:
                            list_down_limit += [el]
                    else:
                        list_down_limit += [self._kwargs_down[kwargs_key][name]]
                    if isinstance(up, list):
                        list_up_limit += up
                    elif isinstance(up, (np.ndarray, jnp.ndarray)):
                        el = up.tolist()
                        if hasattr(el, '__len__'):
                            list_up_limit += el
                        else:
                            list_up_limit += [el]
                    else:
                        list_up_limit += [self._kwargs_up[kwargs_key][name]]
                        

            return (jnp.array(list_down_limit).flatten(),
                    jnp.array(list_up_limit).flatten())

    def update_kwargs(self, kwargs_init=None, kwargs_fixed=None, kwargs_up=None,
                      kwargs_down=None):

        """Updates the kwargs with provided values."""
        if kwargs_init is not None:
            self._kwargs_init = kwargs_init
        if kwargs_fixed is not None:
            self._kwargs_fixed = kwargs_fixed
            self._kwargs_free_indices = self.get_nan_indices()
        if kwargs_init is not None:
            self._kwargs_up = kwargs_up
        if kwargs_init is not None:
            self._kwargs_down = kwargs_down

    def _param_names(self, kwargs, kwargs_key):
        """Setting the parameters."""
        names = []
        kwargs_profile = kwargs[kwargs_key]
        kwargs_fixed_k = self._kwargs_fixed[kwargs_key]
        param_names = self.get_param_names_for_model(kwargs_key)

        for name in param_names:
            if name not in kwargs_fixed_k:
                value = kwargs_profile[name]
                if isinstance(value, (float, int)):
                    names.append(name)
                elif len(value) == 1 and isinstance(value, (list, np.ndarray, np.generic, jnp.ndarray, jnp.generic)):
                    names.append(name)
                elif isinstance(value, (list, np.ndarray, np.generic, jnp.ndarray, jnp.generic)):
                    names += [f'{name}_{i}' for i in range(len(value))]
                else:
                    names.append(name)
            else:
                num_free_param = len(self._kwargs_free_indices[kwargs_key][name])
                if num_free_param == 1:
                    names.append(f'{name}_{self._kwargs_free_indices[kwargs_key][name]}')
                else:
                    names += [f'{name}_{i}' for i in self._kwargs_free_indices[kwargs_key][name]]

        return names

    def get_nan_indices(self):
        kwargs_indice_free = {}
        for kwargs_key in self._kwargs_fixed.keys():
            kwargs_indice_free[kwargs_key] = {}
            for key in self._kwargs_fixed[kwargs_key]:
                indices = jnp.where(jnp.isnan(self._kwargs_fixed[kwargs_key][key]))[0]
                kwargs_indice_free[kwargs_key][key] = indices

        return kwargs_indice_free





import sys
from distutils.core import setup
from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)

try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()

setup(
    name='starred-astro',
    version='1.3',
    author='Kevin Michalewicz, Martin Millon, Fred Dux',
    author_email='kevinmicha@hotmail.com',
    description='A two-channel deconvolution method with Starlet regularization',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['starred', 'starred.deconvolution', 'starred.optim', 'starred.plots', 'starred.psf', 'starred.utils',
              'starred.procedures'],
    requires=['astropy', 'dill', 'jax', 'jaxlib', 'jaxopt', 'matplotlib', 'numpy', 'scipy', 'optax', 'tqdm', 'emcee',
              'pyregion', 'h5py'],
    cmdclass={'test': PyTest}
)



import numpy as np
from copy import deepcopy
import warnings

from starred.psf.psf import PSF
from starred.psf.loss import Loss
from starred.optim.optimization import Optimizer
from starred.psf.parameters import ParametersPSF
from starred.utils.noise_utils import propagate_noise


def build_psf(image, noisemap, subsampling_factor):
    """
    
    Routine taking in cutouts of stars (shape (N, nx, ny), with N the number of star cutouts, and nx,ny the shape of each cutout)
    and their noisemaps (same shape), producing a narrow PSF with pixel grid of the given subsampling_factor

    Parameters
    ----------
    image : array, shape (imageno, nx, ny)
        array containing the data
    noisemap : array, shape (imageno, nx, ny)
        array containing the noisemaps.
    subsampling_factor : int
        by how much we supersample the PSF pixel grid compare to data..

    Returns
    -------
    result : dictionary
        dictionary containing the narrow PSF (key narrow_psf) and other useful things.

    """
    
    # normalize by max of data(numerical precision best with scale ~ 1)
    norm = image.max()
    image /= norm
    noisemap /= norm
    



    model = PSF(image_size=image[0].shape[0], number_of_sources=len(image), 
                upsampling_factor=subsampling_factor, 
                convolution_method='fft',
                include_moffat=True,
                elliptical_moffat=True)
      
    
    smartguess = lambda im: model.smart_guess(im, fixed_background=True)

    # Parameter initialization. 
    
    kwargs_init, kwargs_fixed, kwargs_up, kwargs_down = smartguess(image)

    # smartguess doesn't know about cosmics, other stars ...
    # so we'll be a bit careful.
    medx0 = np.median(kwargs_init['kwargs_gaussian']['x0'])
    medy0 = np.median(kwargs_init['kwargs_gaussian']['y0'])
    kwargs_init['kwargs_gaussian']['x0'] = medx0 * np.ones_like(kwargs_init['kwargs_gaussian']['x0'])
    kwargs_init['kwargs_gaussian']['y0'] = medy0 * np.ones_like(kwargs_init['kwargs_gaussian']['y0'])

    parameters = ParametersPSF(kwargs_init, 
                               kwargs_fixed, 
                               kwargs_up=kwargs_up, 
                               kwargs_down=kwargs_down)

    loss = Loss(image, model, parameters, noisemap**2, len(image),
                regularization_terms='l1_starlet', 
                regularization_strength_scales=0, 
                regularization_strength_hf=0)

    optim = Optimizer(loss, 
                      parameters, 
                      method='l-bfgs-b')
    

    # fit the moffat:
    best_fit, logL_best_fit, extra_fields, runtime = optim.minimize(maxiter=40,
                                                                    restart_from_init=True)

    kwargs_partial = parameters.args2kwargs(best_fit)

    # now moving on to the background.
    # Release backgound, fix the moffat
    kwargs_fixed = {
        'kwargs_moffat': {'fwhm_x': kwargs_partial['kwargs_moffat']['fwhm_x'], 
                          'fwhm_y': kwargs_partial['kwargs_moffat']['fwhm_y'],
                          'phi': kwargs_partial['kwargs_moffat']['phi'],
                          'beta': kwargs_partial['kwargs_moffat']['beta'], 
                          'C': kwargs_partial['kwargs_moffat']['C']},
        'kwargs_gaussian': {},
        'kwargs_background': {},
    }

    parametersfull = ParametersPSF(kwargs_partial,
                                   kwargs_fixed, 
                                   kwargs_up, 
                                   kwargs_down)

    topass = np.nanmedian(noisemap, axis=0)
    topass = np.expand_dims(topass, (0,))
    W = propagate_noise(model, topass, kwargs_init, 
                        wavelet_type_list=['starlet'], 
                        method='MC', num_samples=100,
                        seed=1, likelihood_type='chi2', 
                        verbose=False, 
                        upsampling_factor=subsampling_factor)[0]
    
    lossfull = Loss(image, model, parametersfull, 
                    noisemap**2, len(image), 
                    regularization_terms='l1_starlet',
                    regularization_strength_scales=1.,
                    regularization_strength_hf=1.,
                    regularization_strength_positivity=0, 
                    W=W, 
                    regularize_full_psf=False)
    

    optimfull = Optimizer(lossfull, parametersfull, method='adabelief')
    
        
    optimiser_optax_option = {
                                'max_iterations':2000, 'min_iterations':None,
                                'init_learning_rate': 1e-4, 'schedule_learning_rate':True,
                                # important: restart_from_init True
                                'restart_from_init':True, 'stop_at_loss_increase':False,
                                'progress_bar':True, 'return_param_history':True
                              }           
    
    best_fit, logL_best_fit, extra_fields2, runtime = optimfull.minimize(**optimiser_optax_option)
    
    kwargs_final = parametersfull.args2kwargs(best_fit)
    
    ###########################################################################
    # book keeping
    narrowpsf = model.get_narrow_psf(**kwargs_final, norm=True)
    fullpsf = model.get_full_psf(**kwargs_final, norm=True)
    numpsf    = model.get_background(kwargs_final['kwargs_background'])
    moffat    = model.get_moffat(kwargs_final['kwargs_moffat'], norm=True)
    fullmodel = np.array([model.model(i, **kwargs_final) for i in range(image.shape[0])])
    residuals = image - fullmodel

    result = {
        'model_instance': model,
        'kwargs_psf': kwargs_final,
        'narrow_psf': narrowpsf,
        'full_psf': fullpsf,
        'numerical_psf': numpsf,
        'moffat': moffat,
        'models': fullmodel,
        'residuals': residuals,
        'analytical_optimizer_extra_fields': extra_fields,
        'adabelief_extra_fields': extra_fields2
    }
    ###########################################################################
    return result


def run_multi_steps_PSF_reconstruction(data, model, parameters, sigma_2, masks=None,
                                       lambda_scales=1., lambda_hf=1., lambda_positivity=0.,
                                       fitting_sequence=[['background'], ['moffat']],
                                       optim_list=['l-bfgs-b', 'adabelief'],
                                       kwargs_optim_list=None,
                                       method_noise='MC', regularize_full_psf=False,
                                       verbose=True):
    """
    A high level function for a custom fitting sequence. Similar to build_psf() but with more options.

    :param data: array containing the observations
    :param model: Point Spread Function (PSF) class from ``starred.psf.psf``
    :param parameters: parameters class from ``starred.psf.parameters``
    :param sigma_2: array containing the square of the noise maps
    :param lambda_scales: Lagrange parameter that weights intermediate scales in the transformed domain.
    :param lambda_hf: Lagrange parameter weighting the highest frequency scale
    :param lambda_positivity: Lagrange parameter weighting the positivity of the full PSF. 0 means no positivity constraint (recommended).
    :param fitting_sequence: list, List of lists, containing the element of the model to keep fixed. Example : [['pts-source-astrometry','pts-source-photometry','background'],['pts-source-astrometry','pts-source-photometry'], ...]
    :param optim_list: List of optimiser. Recommended if background is kept constant : 'l-bfgs-b', 'adabelief' otherwise.
    :param kwargs_optim_list: List of dictionary, containing the setting for the different optimiser.
    :param method_noise: method for noise propagation. Choose 'MC' for an empirical propagation of the noise or 'SLIT' for analytical propagation.
    :param regularize_full_psf: True if you want to regularize the Moffat and the background (recommended). False regularizes only the background
    :param masks: array containing the masks for the PSF (if given)

    :return model, parameters, loss, kwargs_partial_list, LogL_list, loss_history_list
    """

    # Check the sequence
    assert len(fitting_sequence) == len(optim_list), "Fitting sequence and optimiser list have different lenght !"
    if kwargs_optim_list is not None:
        assert len(fitting_sequence) == len(
            kwargs_optim_list), "Fitting sequence and kwargs optimiser list have different lenght !"
    else:
        warnings.warn('No optimiser kwargs provided. Default configuration is used.')
        kwargs_optim_list = [{} for _ in range(len(fitting_sequence))]
    kwargs_init, kwargs_fixed_default, kwargs_up, kwargs_down = deepcopy(parameters._kwargs_init), deepcopy(
        parameters._kwargs_fixed), \
        deepcopy(parameters._kwargs_up), deepcopy(parameters._kwargs_down)

    kwargs_partial_list = [kwargs_init]
    loss_history_list = []
    LogL_list = []
    W = None

    for i, steps in enumerate(fitting_sequence):
        kwargs_fixed = deepcopy(kwargs_fixed_default)
        background_free = True
        print(f'### Step {i + 1}, fixing : {steps} ###')
        for fixed_feature in steps:
            if fixed_feature == 'pts-source-astrometry':
                kwargs_fixed['kwargs_gaussian']['x0'] = kwargs_partial_list[i]['kwargs_gaussian']['x0']
                kwargs_fixed['kwargs_gaussian']['y0'] = kwargs_partial_list[i]['kwargs_gaussian']['y0']
            elif fixed_feature == 'pts-source-photometry':
                kwargs_fixed['kwargs_gaussian']['a'] = kwargs_partial_list[i]['kwargs_gaussian']['a']
                kwargs_fixed['kwargs_moffat']['C'] = kwargs_partial_list[i]['kwargs_moffat']['C']
            elif fixed_feature == 'background':
                # TODO: check if there is a speed up when skipping regularisation in the case of a fixed background
                kwargs_fixed['kwargs_background']['background'] = kwargs_partial_list[i]['kwargs_background'][
                    'background']
                background_free = False
            elif fixed_feature == 'moffat':
                if model.elliptical_moffat:
                    kwargs_fixed['kwargs_moffat']['fwhm_x'] = kwargs_partial_list[i]['kwargs_moffat']['fwhm_x']
                    kwargs_fixed['kwargs_moffat']['fwhm_y'] = kwargs_partial_list[i]['kwargs_moffat']['fwhm_y']
                    kwargs_fixed['kwargs_moffat']['phi'] = kwargs_partial_list[i]['kwargs_moffat']['phi']
                else:
                    kwargs_fixed['kwargs_moffat']['fwhm'] = kwargs_partial_list[i]['kwargs_moffat']['fwhm']
                kwargs_fixed['kwargs_moffat']['beta'] = kwargs_partial_list[i]['kwargs_moffat']['beta']
                kwargs_fixed['kwargs_moffat']['C'] = kwargs_partial_list[i]['kwargs_moffat']['C']
            else:
                raise ValueError(
                    f'Steps {steps} is not defined. Choose between "pts-source-astrometry", "pts-source-photometry", "background" or "moffat"')

        # Lift degeneracy between background and Moffat by fixing Moffat amplitude
        if background_free:
            kwargs_fixed['kwargs_moffat']['C'] = kwargs_partial_list[i]['kwargs_moffat']['C']
            lambda_scales_eff = deepcopy(lambda_scales)
            lambda_hf_eff = deepcopy(lambda_hf)
        else:
            # remove regularisation for speed up
            lambda_scales_eff = 0.
            lambda_hf_eff = 0.

        # recompile the parameter class as we have changed the number of free parameters
        parameters = ParametersPSF(kwargs_partial_list[i], kwargs_fixed, kwargs_up, kwargs_down)
        loss = Loss(data, model, parameters, sigma_2, model.M, masks=masks, regularization_terms='l1_starlet',
                    regularization_strength_scales=lambda_scales_eff, regularization_strength_hf=lambda_hf_eff,
                    regularization_strength_positivity=lambda_positivity, W=W,
                    regularize_full_psf=regularize_full_psf)

        optim = Optimizer(loss, parameters, method=optim_list[i])
        best_fit, logL_best_fit, extra_fields, runtime = optim.minimize(**kwargs_optim_list[i])
        if verbose:
            try:
                # this will work only for the Jaxopt optimiser, which have a success argument
                if extra_fields['stat'].success:
                    print(
                        f'Success of the step {i + 1} fit in {extra_fields["stat"].iter_num} iterations ({runtime} s)')
                else:
                    print(f'Warning: step {i + 1} fit did not converge !')
            except:
                pass

        # Saving partial results
        kwargs_partial_steps = deepcopy(parameters.best_fit_values(as_kwargs=True))
        loss_history_list.append(extra_fields['loss_history'])
        LogL_list.append(logL_best_fit)

        # compute noise propagation
        W = propagate_noise(model, np.sqrt(sigma_2), kwargs_partial_steps, wavelet_type_list=['starlet'],
                            method=method_noise,
                            num_samples=400, seed=1, likelihood_type='chi2', verbose=False,
                            upsampling_factor=model.upsampling_factor)[0]

        # update kwargs_partial_list
        kwargs_partial_list.append(deepcopy(kwargs_partial_steps))
        if verbose:
            print('Step %i/%i took %2.f seconds' % (i + 1, len(fitting_sequence), runtime))
            print('Kwargs partial at step %i/%i' % (i + 1, len(fitting_sequence)), kwargs_partial_steps)
            print('LogL : ', logL_best_fit)
            print('Overall Reduced Chi2 : ',
                  -2 * loss._log_likelihood_chi2(kwargs_partial_steps) / (model.image_size ** 2))

    return model, parameters, loss, kwargs_partial_list, LogL_list, loss_history_list



import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from matplotlib.widgets import Slider
from copy import deepcopy
from astropy.visualization import simple_norm

from starred.utils.generic_utils import Downsample


def single_PSF_plot(model, data, sigma_2, kwargs, n_psf=0, figsize=(15, 8), units=None, upsampling=None, masks=None):
    """
    Plots the narrow PSF fit for a single observation.

    :param model: array containing the model
    :param data: array containing the observations
    :param sigma_2: array containing the square of the noise maps
    :param kwargs: dictionary containing the parameters of the model
    :param n_psf: selected PSF index
    :type n_psf: int
    :param figsize: tuple that indicates the size of the figure
    :param units: units in which the pixel values are expressed
    :type units: str
    :param upsampling_factor: Provide the upsampling factor to degrade the model to the image resolution. Leave to 'None' to show the higher resolution model.
    :type upsampling_factor: int
    :param masks: Boolean masks
    :type masks: array of the size of your image

    :return: output figure

    """
    if units is not None:
        str_unit = '[' + units + ']'
    else:
        str_unit = ''

    if masks is not None:
        alphas = deepcopy(masks)
        ind = np.where(masks == 0)
        alphas[ind] = 0.7
    else:
        alphas = np.ones_like(data)

    estimated_full_psf = model.model(n_psf, **kwargs)
    analytic = model.get_moffat(kwargs['kwargs_moffat'], norm=True)
    s = model.get_narrow_psf(**kwargs, norm=True)
    background = model.get_background(kwargs['kwargs_background'])

    if upsampling is not None:
        analytic = Downsample(analytic, factor=upsampling)
        s = Downsample(s, factor=upsampling)
        background = Downsample(background, factor=upsampling)

    dif = data[n_psf, :, :] - estimated_full_psf
    rr = dif / np.sqrt(sigma_2[n_psf, :, :])

    fig, axs = plt.subplots(2, 3, figsize=figsize)
    fraction = 0.046
    pad = 0.04
    font_size = 14
    ticks_size = 6

    plt.rc('font', size=font_size)
    axs[0, 0].set_title('Data %s' % str_unit, fontsize=font_size)
    axs[0, 0].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[0, 1].set_title('PSF model %s' % str_unit, fontsize=font_size)
    axs[0, 1].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[0, 2].set_title('Map of relative residuals', fontsize=font_size)
    axs[0, 2].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[1, 0].set_title('Moffat', fontsize=font_size)
    axs[1, 0].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[1, 1].set_title('Grid of pixels', fontsize=font_size)
    axs[1, 1].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[1, 2].set_title('Narrow PSF', fontsize=font_size)
    axs[1, 2].tick_params(axis='both', which='major', labelsize=ticks_size)

    fig.colorbar(axs[0, 0].imshow(data[n_psf, :, :], norm=colors.SymLogNorm(linthresh=100), origin='lower'),
                 ax=axs[0, 0], fraction=fraction, pad=pad, format='%.0e')
    fig.colorbar(axs[0, 1].imshow(estimated_full_psf, norm=colors.SymLogNorm(linthresh=100), origin='lower'),
                 ax=axs[0, 1], fraction=fraction, pad=pad, format='%.0e')
    fig.colorbar(axs[0, 2].imshow(rr, origin='lower', alpha = alphas[n_psf,:,:]), ax=axs[0, 2], fraction=fraction, pad=pad)
    fig.colorbar(axs[1, 0].imshow(analytic, norm=colors.SymLogNorm(linthresh=1e-2), origin='lower'), ax=axs[1, 0],
                 fraction=fraction,
                 pad=pad)
    fig.colorbar(axs[1, 1].imshow(background, origin='lower'), ax=axs[1, 1], fraction=fraction,
                 pad=pad)
    fig.colorbar(axs[1, 2].imshow(s, norm=colors.SymLogNorm(linthresh=1e-3), origin='lower'), ax=axs[1, 2],
                 fraction=fraction, pad=pad)

    for ax in np.array(axs).flatten():
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

    plt.tight_layout()

    return fig


def multiple_PSF_plot(model, data, sigma_2, kwargs, masks=None, figsize=None, units=None, vmin=None, vmax=None):
    """
    Plots the narrow PSF fit for all observations.

    :param model: array containing the model
    :param data: array containing the observations
    :param sigma_2: array containing the square of the noise maps
    :param kwargs: dictionary containing the parameters of the model
    :param masks:  array containing the masks
    :param figsize: tuple that indicates the size of the figure
    :param units: units in which the pixel values are expressed
    :type units: str
    :param vmin: lower limit for displaying the residuals (in unit of noise sigma)
    :param vmax: upper limit for displaying the residuals (in unit of noise sigma)

    :return: output figure
    """
    if units is not None:
        str_unit = '[' + units + ']'
    else:
        str_unit = ''

    if figsize is None:
        nimage,nx,ny = np.shape(data)
        figsize = (12+nimage*2, 10)
    fig, axs = plt.subplots(2, model.M, figsize=figsize)
    plt.subplots_adjust(wspace=0.3)
    if model.M == 1:
        axs = np.asarray([axs]).T
    fraction = 0.046
    pad = 0.04
    font_size = 14
    plt.rc('font', size=12)
    fmt_PSF = '%.0e'
    fmt_residuals = '%2.f'

    if masks is not None:
        alphas = deepcopy(masks)
        ind = np.where(masks == 0)
        alphas[ind] = 0.7
        kargs = [{'alpha':alphas[i,:,:]} for i in range(model.M)]
    else:
        kargs = [{} for i in range(model.M)]

    for ka in kargs: 
        if vmin is not None:
            ka['vmin'] = vmin
        if vmax is not None:
            ka['vmax'] = vmax

    for i in range(model.M):
        estimated_full_psf = model.model(i, **kwargs)
        axs[0, i].set_title('PSF model %i %s' % (i + 1, str_unit), fontsize=font_size)
        axs[0, i].tick_params(axis='both', which='major', labelsize=10)
        axs[1, i].set_title('Relative residuals %i' % (i + 1), fontsize=font_size)
        axs[1, i].tick_params(axis='both', which='major', labelsize=10)

        fig.colorbar(axs[0, i].imshow(estimated_full_psf, norm=colors.SymLogNorm(linthresh=100), origin='lower'),
                     ax=axs[0, i], fraction=fraction, pad=pad, format=fmt_PSF)
        fig.colorbar(axs[1, i].imshow((data[i, :, :] - estimated_full_psf) / np.sqrt(sigma_2[i, :, :]),
                                      origin='lower', **kargs[i]), ax=axs[1, i], fraction=fraction, pad=pad, format=fmt_residuals)

    for ax in np.array(axs).flatten():
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

    return fig


def display_data(data, sigma_2=None, masks=None, figsize=None, units=None, center=None):
    """
    Plots the observations and the noise maps.

    :param data: array containing the observations
    :param sigma_2: array containing the square of the noise maps
    :param figsize: tuple that indicates the size of the figure
    :param units: units in which the pixel values are expressed
    :type units: str
    :param center: x and y coordinates of the centers of the observations

    :return: output figure
    """
    if sigma_2 is None:
        row = 1
        show_sigma = False
    else:
        row = 2
        show_sigma = True

    if figsize is None:
        nimage,nx,ny = np.shape(data)
        figsize = (12+nimage*2, 10)

    if units is not None:
        str_unit = '[' + units + ']'
    else:
        str_unit = ''

    n_image, nx, ny = np.shape(np.asarray(data))
    fig, axs = plt.subplots(row, n_image, figsize=figsize)
    plt.subplots_adjust(wspace=0.3)
    if row == 1 and n_image == 1:
        axs = np.asarray([[axs]])
    elif row == 1:
        axs = np.asarray([axs])
    elif n_image == 1:
        axs = np.asarray([axs]).T
    fraction = 0.046
    pad = 0.04
    fontsize = 12

    if masks is not None:
        alphas = deepcopy(masks)
        ind = np.where(masks == 0)
        alphas[ind] = 0.9
        kargs = [{'alpha':alphas[i,:,:]} for i in range(n_image)]
    else:
        kargs = [{} for i in range(n_image)]

    for i in range(n_image):
        plt.rc('font', size=12)
        axs[0][i].set_title('Data %i %s' % (i + 1, str_unit), fontsize=fontsize)
        axs[0][i].tick_params(axis='both', which='major', labelsize=10)

        if show_sigma:
            axs[1][i].set_title('Noise map %i %s' % (i + 1, str_unit), fontsize=fontsize)
            axs[1][i].tick_params(axis='both', which='major', labelsize=10)

        fig.colorbar(axs[0][i].imshow(data[i, :, :], norm=colors.SymLogNorm(linthresh=10), origin='lower',  **kargs[i]),
                     ax=axs[0][i],
                     fraction=fraction, pad=pad, format='%.0e')
        if center is not None:
            c_x, c_y = center[0], center[1]
            axs[0][i].scatter(nx / 2. + c_x[i] - 0.5, ny / 2. + c_y[i] - 0.5, marker='x',
                              c='r')  # +0.5 to mach matplotplit pixel convention

        if show_sigma:
            fig.colorbar(
                axs[1][i].imshow(np.sqrt(sigma_2[i, :, :]), norm=colors.SymLogNorm(linthresh=10), origin='lower', **kargs[i]),
                ax=axs[1][i],
                fraction=fraction, pad=pad, format='%2.f')

    for ax in axs.flatten():
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    plt.tight_layout()

    return fig

def dict_to_kwargs_list(dict):
        """
        Transform dictionnary into a list kwargs. All entry must have the same lenght.

        :param
        """
        k_list = []
        keys = list(dict.keys())
        for i in range(len(dict[keys[0]])):
            k_list.append({})
            for key in keys:
                k_list[i][key]=dict[key][i]

        return k_list
    
def plot_deconvolution(model, data, sigma_2, s, kwargs, epoch=0, units=None, figsize=(15, 10), cut_dict=None):
    """
    Plots the results of the deconvolution.

    :param data: array containing the observations. Has shape (n_epoch, n_pixel, n_pixel).
    :param sigma_2: array containing the square of the noise maps (n_epoch, n_pixel, n_pixel).
    :param s: array containing the narrow PSF (n_epoch, n_pixel*susampling factor, n_pixel*susampling factor).
    :param epoch: index of the epoch to plot
    :param kwargs: dictionary containing the parameters of the model
    :type epoch: int
    :param figsize: tuple that indicates the size of the figure
    :param units: units in which the pixel values are expressed
    :type units: str

    :return: output figure
    """

    if units is not None:
        str_unit = '[' + units + ']'
    else:
        str_unit = ''

    if cut_dict is None :
        # Default setting
        cut_dict = {
            'linthresh':[5e2,5e2,None,5e1,5e1,1e-3],
            'vmin':[None, None, None, None, None, None],
            'vmax':[None, None, None, None, None, None],
        }

    k_dict = dict_to_kwargs_list(cut_dict)
    output = model.model(kwargs)[epoch]
    deconv, h = model.getDeconvolved(kwargs, epoch)
    data_show = data[epoch, :, :]

    dif = data_show - output
    rr = np.abs(dif) / np.sqrt(sigma_2[epoch, :, :])

    fig, axs = plt.subplots(2, 3, figsize=(15, 8))
    fraction = 0.046
    pad = 0.04
    font_size = 10
    ticks_size = 6

    plt.rc('font', size=font_size)
    axs[0, 0].set_title(f'Data {str_unit}', fontsize=8)
    axs[0, 0].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[0, 1].set_title(f'Convolving back {str_unit}', fontsize=8)
    axs[0, 1].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[0, 2].set_title('Map of relative residuals', fontsize=8)
    axs[0, 2].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[1, 0].set_title(f'Background {str_unit}', fontsize=8)
    axs[1, 0].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[1, 1].set_title(f'Deconvolved image {str_unit}', fontsize=8)
    axs[1, 1].tick_params(axis='both', which='major', labelsize=ticks_size)
    axs[1, 2].set_title('Narrow PSF', fontsize=8)
    axs[1, 2].tick_params(axis='both', which='major', labelsize=ticks_size)

    fig.colorbar(axs[0, 0].imshow(data_show, norm=colors.SymLogNorm(**k_dict[0]), origin='lower'), ax=axs[0, 0], fraction=fraction, pad=pad)
    fig.colorbar(axs[0, 1].imshow(output, norm=colors.SymLogNorm(**k_dict[1]), origin='lower'), ax=axs[0, 1], fraction=fraction,pad=pad)
    if 'linthresh' in k_dict[2].keys():
        del k_dict[2]['linthresh']
    fig.colorbar(axs[0, 2].imshow(rr, origin='lower', **k_dict[2]), ax=axs[0, 2], fraction=fraction, pad=pad)
    fig.colorbar(axs[1, 0].imshow(h, norm=colors.SymLogNorm(**k_dict[3]), origin='lower'), ax=axs[1, 0], fraction=fraction,pad=pad)
    fig.colorbar(axs[1, 1].imshow(deconv, norm=colors.SymLogNorm(**k_dict[4]), origin='lower'), ax=axs[1, 1],fraction=fraction, pad=pad)
    fig.colorbar(axs[1, 2].imshow(s[epoch, :, :], norm=colors.SymLogNorm(**k_dict[5]), origin='lower'), ax=axs[1, 2],fraction=fraction, pad=pad)

    return fig


def view_deconv_model(model, kwargs, data, sigma_2, figsize=(9, 7.5), cmap='gist_heat'):
    output = model.model(kwargs)
    psf = model.psf
    noisemap = sigma_2 ** 0.5

    # setup for first epoch
    deconvs = [model.getDeconvolved(kwargs, i) for i in range(len(output))]
    decs, hs = zip(*deconvs)
    # subtract the constant component from h
    hs = [h - kwargs['kwargs_background']['mean'][i] for i, h in enumerate(hs)]
    deconv = decs[0]
    h = hs[0]
    
    normdeconv = simple_norm(deconv, stretch='asinh', percent=99.9)
    if np.std(h) > 0.:
        normh = simple_norm(h, stretch='asinh', percent=99.99)
    else:
        normh = simple_norm(h)
        normh.vmin = 0
        normh.vmax = 1

    s = psf[0]

    ##########################################################################
    # figure
    fig, axs = plt.subplots(2, 3, figsize=figsize)
    for ax in axs.flatten():
        ax.set_xticks([])
        ax.set_yticks([])

    datap = axs[0, 0].imshow(data[0], origin='lower', cmap=cmap)
    axs[0, 0].set_title('data')

    modelp = axs[0, 1].imshow(output[0], origin='lower', cmap=cmap)
    axs[0, 1].set_title('model')

    diffp = axs[1, 0].imshow((data[0] - output[0]) / noisemap[0], origin='lower', cmap=cmap)
    axs[1, 0].set_title('(data-model)/noise')

    backp = axs[1, 1].imshow(h, origin='lower', cmap=cmap, norm=normh)
    axs[1, 1].set_title('background')

    decp = axs[0, 2].imshow(deconv, origin='lower', cmap=cmap, norm=normdeconv)
    axs[0, 2].set_title('deconvolved')

    psfp = axs[1, 2].imshow(s, origin='lower', cmap=cmap)
    axs[1, 2].set_title('narrow psf')

    plt.tight_layout()

    if len(output)>1:
        axcolor   = 'lightgoldenrodyellow'
        axslider  = plt.axes([0.1, 0.05, 0.75, 0.01], facecolor=axcolor)
        slider    = Slider(axslider, 'Epoch', 0, len(output)-1, valinit=0, valstep=1)
        #######################################################################
        # functions for slider update, only if more than one epoch.
        def press(event): #pragma: no cover
            try:
                button = event.button
            except:
                button = 'None'
            if event.key == 'right' or button == 'down':
                if slider.val < len(output) - 1:
                    slider.set_val(slider.val + 1)
            elif event.key == 'left' or button == 'up':
                if slider.val > 0:
                    slider.set_val(slider.val - 1)
            update(slider.val)
            fig.canvas.draw_idle()
        
        def reset(event):#pragma: no cover
            slider.reset()
            
        def update(val):#pragma: no cover
            epoch0 = int(slider.val)
            deconv, h = decs[epoch0], hs[epoch0]
            s = psf[epoch0]
            # update all the plots
            datap.set_data(data[epoch0])
            modelp.set_data(output[epoch0])
            diffp.set_data((data[epoch0] - output[epoch0])/noisemap[epoch0])
            backp.set_data(h)
            decp.set_data(deconv)
            psfp.set_data(s)
            
    
        fig.canvas.mpl_connect('key_press_event', press)
        fig.canvas.mpl_connect('scroll_event', press)
        slider.on_changed(update)
    
    plt.show(block=False)


def make_movie(model, kwargs, data, sigma_2, outpath, figsize=(9, 7.5), epochs_list=None, duration=20, loop=1,
               format='gif', cmap=None):
    output = model.model(kwargs)
    psf = model.psf
    noisemap = sigma_2 ** 0.5

    if epochs_list is None:
        epochs_list = range(len(output))

    # setup for first epoch
    deconvs = [model.getDeconvolved(kwargs, i) for i in range(len(output))]
    deconv, h = deconvs[0]
    s = psf[0]

    ##########################################################################
    # figure
    fig, axs = plt.subplots(2, 3, figsize=figsize)
    for ax in axs.flatten():
        ax.set_xticks([])
        ax.set_yticks([])

    datap = axs[0, 0].imshow(data[0], origin='lower', cmap=cmap)
    axs[0, 0].set_title('data')

    modelp = axs[0, 1].imshow(output[0], origin='lower', cmap=cmap)
    axs[0, 1].set_title('model')

    diffp = axs[1, 0].imshow((data[0] - output[0]) / noisemap[0], origin='lower', cmap=cmap)
    axs[1, 0].set_title('(data-model)/noise')

    backp = axs[1, 1].imshow(h, origin='lower', cmap=cmap)
    axs[1, 1].set_title('background')

    decp = axs[0, 2].imshow(deconv, origin='lower', cmap=cmap)
    axs[0, 2].set_title('deconvolved')

    psfp = axs[1, 2].imshow(s, origin='lower', cmap=cmap)
    axs[1, 2].set_title('narrow psf')

    plt.tight_layout()

    # update all the plots
    files = []
    for i, epoch0 in enumerate(epochs_list):
        deconv, h = deconvs[epoch0]
        s = psf[epoch0]
        datap.set_data(data[epoch0])
        modelp.set_data(output[epoch0])
        diffp.set_data((data[epoch0] - output[epoch0]) / noisemap[epoch0])
        backp.set_data(h)
        decp.set_data(deconv)
        psfp.set_data(s)

        file_png = os.path.join(outpath, "frame{0:05d}.png".format(i))
        fig.savefig(file_png)
        files.append(file_png)

    if format == 'gif':
        gif_name = os.path.join(outpath, "deconv.gif")
        make_gif(files, gif_name, duration=duration, loop=loop)
    elif format == 'mp4v':  # pragma: no cover
        video_name = os.path.join(outpath, f'deconv.{format}')
        fps = len(files) / duration
        make_video(files, outvid=video_name, fps=fps, size=None,
                   is_color=True, format=format)
    else:
        RuntimeError('Unsupported video format. Use "gif" or "mp4v".')


def make_gif(list_files, output_path, duration=100, loop= 1):
    try:
        from PIL import Image
    except ImportError as e:
        print(e)
        print('Python package PIL is required for gif creation.')
    frames = [Image.open(image) for image in list_files]
    frame_one = frames[0]
    frame_one.save(output_path, format="GIF", append_images=frames,
               save_all=True, duration=duration)

def make_video(images, outvid=None, fps=5, size=None,
               is_color=True, format="mp4v"):  # pragma: no cover
    """
    Create a video from a list of images.

    @param      outvid      output video
    @param      images      list of images to use in the video
    @param      fps         frame per second
    @param      size        size of each frame
    @param      is_color    color
    @param      format      see http://www.fourcc.org/codecs.php
    @return                 see http://opencv-python-tutroals.readthedocs.org/en/latest/py_tutorials/py_gui/py_video_display/py_video_display.html

    The function relies on http://opencv-python-tutroals.readthedocs.org/en/latest/.
    By default, the video will have the size of the first image.
    It will resize every image to this size before adding them to the video.
    """
    try:
        from cv2 import VideoWriter, VideoWriter_fourcc, imread, resize
    except ImportError as e:
        print(e)
        print('Python package opencv-python is required for video creation.')

    fourcc = VideoWriter_fourcc(*format)
    vid = None
    for image in images:
        if not os.path.exists(image):
            raise FileNotFoundError(image)
        img = imread(image)
        if vid is None:
            if size is None:
                size = img.shape[1], img.shape[0]
            vid = VideoWriter(outvid, fourcc, float(fps), size, is_color)
        if size[0] != img.shape[1] and size[1] != img.shape[0]:
            img = resize(img, size)
        vid.write(img)
    vid.release()
    return vid

def plot_loss(loss_history, figsize = (10,5),  ax = None, title = None):
    if ax is None:
        fig, ax = plt.subplots(1,1, figsize =figsize)
    ax.plot(range(len(loss_history)), loss_history)
    ax.set_xlabel('Steps')
    ax.set_ylabel('Loss')
    ax.set_yscale('log')

    if title is not None: 
        ax.set_title(title)

    return fig

def plot_convergence_by_walker(samples_mcmc, param_mcmc, n_walkers, verbose = False):
    n_params = samples_mcmc.shape[1]
    n_step = int(samples_mcmc.shape[0] / n_walkers)

    chain = np.empty((n_walkers, n_step, n_params))

    for i in np.arange(n_params):
        samples = samples_mcmc[:, i].T
        chain[:, :, i] = samples.reshape((n_step, n_walkers)).T

    mean_pos = np.zeros((n_params, n_step))
    median_pos = np.zeros((n_params, n_step))
    std_pos = np.zeros((n_params, n_step))
    q16_pos = np.zeros((n_params, n_step))
    q84_pos = np.zeros((n_params, n_step))

    # chain = np.empty((nwalker, nstep, ndim), dtype = np.double)
    for i in np.arange(n_params):
        for j in np.arange(n_step):
            mean_pos[i][j] = np.mean(chain[:, j, i])
            median_pos[i][j] = np.median(chain[:, j, i])
            std_pos[i][j] = np.std(chain[:, j, i])
            q16_pos[i][j] = np.percentile(chain[:, j, i], 16.)
            q84_pos[i][j] = np.percentile(chain[:, j, i], 84.)

    fig, ax = plt.subplots(n_params, sharex=True, figsize=(16, 2 * n_params))
    if n_params == 1: ax = [ax]
    last = n_step
    burnin = int((9.*n_step) / 10.) #get the final value on the last 10% on the chain

    for i in range(n_params):
        if verbose :
            print(param_mcmc[i], '{:.4f} +/- {:.4f}'.format(median_pos[i][last - 1], (q84_pos[i][last - 1] - q16_pos[i][last - 1]) / 2))
        ax[i].plot(median_pos[i][:last], c='g')
        ax[i].axhline(np.median(median_pos[i][burnin:last]), c='r', lw=1)
        ax[i].fill_between(np.arange(last), q84_pos[i][:last], q16_pos[i][:last], alpha=0.4)
        ax[i].set_ylabel(param_mcmc[i], fontsize=10)
        ax[i].set_xlim(0, last)

    return fig
